# Copyright 2023 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from tempfile import NamedTemporaryFile

from portmodlib.portmod import Group, Person, parse_package_metadata


def test_metadata():
    with NamedTemporaryFile("w", delete=False) as file:
        print(
            """
maintainer: me@example.org
use:
    test: description
    other: foo
upstream:
    maintainer: Someone <foo@example.org>
""",
            file=file,
        )

    metadata = parse_package_metadata(file.name)

    assert metadata.use == {"test": "description", "other": "foo"}
    assert isinstance(metadata.maintainer, Person)
    # Not recognized as the email if there is just an email
    assert str(metadata.maintainer) == metadata.maintainer.name == "me@example.org"

    assert metadata.upstream
    assert isinstance(metadata.upstream.maintainer, Person)
    assert str(metadata.upstream.maintainer) == "Someone <foo@example.org>"
    assert metadata.upstream.maintainer.name == "Someone"
    assert metadata.upstream.maintainer.email == "foo@example.org"
    os.remove(file.name)


def test_metadata_maintainer_nonstring():
    with NamedTemporaryFile("w", delete=False) as file:
        print(
            """
maintainer:
    name: Someone
    email: foo@example.org
    desc: This is a test

upstream:
    maintainer:
    - name: Someone
    - Someone else
""",
            file=file,
        )

    metadata = parse_package_metadata(file.name)

    assert isinstance(metadata.maintainer, Person)
    assert metadata.maintainer.name == "Someone"
    assert metadata.maintainer.email == "foo@example.org"
    assert metadata.maintainer.desc == "This is a test"
    assert str(metadata.maintainer) == "Someone <foo@example.org>"

    assert metadata.upstream
    assert isinstance(metadata.upstream.maintainer, list)
    assert all(isinstance(person, Person) for person in metadata.upstream.maintainer)
    assert list(map(str, metadata.upstream.maintainer)) == ["Someone", "Someone else"]
    os.remove(file.name)


def test_metadata_maintainer_group():
    with NamedTemporaryFile("w", delete=False) as file:
        print(
            """
maintainer:
    group: Some group
""",
            file=file,
        )

    metadata = parse_package_metadata(file.name)

    assert isinstance(metadata.maintainer, Group)
    assert metadata.maintainer.group == "Some group"
    os.remove(file.name)



from typing import Any, Dict, List, Union

from PySide6.QtCore import (
    QAbstractTableModel,
    QModelIndex,
    QPersistentModelIndex,
    QSortFilterProxyModel,
    Qt,
    Signal,
    Slot,
)

from portmod._gui.human_bytes import format_bytes
from portmod._gui.Manage.FlagListModel import FlagListModel
from portmod._gui.packages import get_installed_packages, get_local_flags
from portmod.download import get_total_download_size
from portmod.pybuild import InstalledPybuild
from portmod.query import get_package_metadata
from portmod.repo import get_repo

# TODO: Maybe _data can be a dictionary with InstalledPybuilds as keys, and the flags as values.

# TODO: The logic behind most of these getters should be moved to the real model. Leave the proxy
# model for just sorting and such.


class InstalledPackagesProxyModel(QSortFilterProxyModel):
    flagModels: Dict[str, FlagListModel] = {}
    sortOrders: List[Union[Qt.SortOrder, None]]
    prefixChanged = Signal()

    def __init__(self, data: List[InstalledPybuild]) -> None:
        super().__init__()

        self.sortOrders = [None, None, None, None]
        self.realModel = InstalledPackagesModel(data)
        self.setSourceModel(self.realModel)

    @Slot(int, result=str)
    def getAtom(self, rowIndex: int) -> str:
        """
        Returns the atom of the given row, and returns it in string form.
        """

        # TODO: Should this be CPN?
        return str(self.get_source_pybuild(rowIndex).CPN)

    @Slot(int, result=str)
    def get_name(self, rowIndex: int) -> str:
        return self.get_source_pybuild(rowIndex).NAME

    @Slot(int, result=str)
    def get_author(self, rowIndex: int) -> str:
        pybuild = self.get_source_pybuild(rowIndex)
        atom = pybuild.CPN
        repo = get_repo(pybuild.REPO)
        metadata = get_package_metadata(atom, repo)

        if metadata:
            return ", ".join(metadata.upstream_maintainers)
        else:
            return ""

    @Slot(int, result=str)
    def get_version(self, rowIndex: int) -> str:
        return str(self.get_source_pybuild(rowIndex).version)

    @Slot(int, result=str)
    def get_size(self, rowIndex: int) -> str:
        size = get_total_download_size([self.get_source_pybuild(rowIndex)])
        return format_bytes(size)

    @Slot(int, result=str)
    def get_license(self, rowIndex: int) -> str:
        _license = self.get_source_pybuild(rowIndex).LICENSE
        if _license:
            return _license
        else:
            return "No license"

    @Slot(int, result=str)
    def get_description(self, rowIndex: int) -> str:
        return self.get_source_pybuild(rowIndex).DESC

    @Slot(int, result=str)
    def get_homepage(self, rowIndex: int) -> str:
        return self.get_source_pybuild(rowIndex).HOMEPAGE

    @Slot(str, result=FlagListModel)
    def get_local_flags_model(self, row_index: int) -> FlagListModel:
        atom = self.get_source_pybuild(row_index).CPN
        model = self.flagModels.get(atom)
        if model:
            return model
        else:
            self.flagModels[atom] = FlagListModel(
                get_local_flags(self.get_source_pybuild(row_index))
            )
            return self.flagModels[atom]

    @Slot()
    def changeToCurrentPrefix(self) -> None:
        self.realModel.changeToCurrentPrefix()
        self.prefixChanged.emit()

    def setSortOrder(self, columnIndex: int, sortOrder: Qt.SortOrder) -> None:
        self.sort(columnIndex, sortOrder)
        self.sortOrders[columnIndex] = sortOrder

    @Slot(int)
    def toggleSort(self, columnIndex: int) -> None:
        """
        Toggle between sort orders at the specified column index. If the sort order is ascending,
        change it to descending. If it's descending order not set, change it to ascending.
        """
        # TODO: Add 🢓 or 🢑 to column header based on sort order
        if self.sortOrders[columnIndex] == Qt.SortOrder.AscendingOrder:
            self.setSortOrder(columnIndex, Qt.SortOrder.DescendingOrder)
        elif (
            self.sortOrders[columnIndex] == Qt.SortOrder.DescendingOrder
            or self.sortOrders[columnIndex] is None
        ):
            self.setSortOrder(columnIndex, Qt.SortOrder.AscendingOrder)

    def get_source_pybuild(self, proxy_row_index: int) -> InstalledPybuild:
        source_row_index = self.mapToSource(self.index(int(proxy_row_index), 0)).row()
        pybuild: InstalledPybuild = self.realModel._data[source_row_index]
        return pybuild


class InstalledPackagesModel(QAbstractTableModel):
    def __init__(self, data: List[InstalledPybuild]):
        super(InstalledPackagesModel, self).__init__()
        self._data = data

    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: Qt.ItemDataRole) -> Any:  # type: ignore
        if role == Qt.ItemDataRole.DisplayRole:
            pybuild = self._data[index.row()]
            column_data = [
                pybuild.PN,
                ", ".join(pybuild.get_use()),
                pybuild.CATEGORY,
                str(pybuild.version),
            ]

            return column_data[index.column()]

    def rowCount(self, index: Union[QModelIndex, QPersistentModelIndex]) -> int:  # type: ignore
        return len(self._data)

    def columnCount(self, index: Union[QModelIndex, QPersistentModelIndex]) -> int:  # type: ignore
        return 4

    def headerData(self, section: int, orientation: Qt.Orientation, role: Qt.ItemDataRole) -> str:  # type: ignore
        header_labels = ["Name", "Use Flags", "Category", "Version"]

        if (
            role == Qt.ItemDataRole.DisplayRole
            and orientation == Qt.Orientation.Horizontal
        ):
            return header_labels[section]

        header_data: str = QAbstractTableModel.headerData(
            self, section, orientation, role
        )
        return header_data

    def changeToCurrentPrefix(self) -> None:
        """
        Repopulates the table with installed packages from the current prefix.
        Used when the prefix is changed while the GUI is still running.
        """
        self.beginRemoveRows(QModelIndex(), 0, len(self._data))
        self._data.clear()
        self.endRemoveRows()

        new_data = get_installed_packages()

        self.beginInsertRows(QModelIndex(), 0, len(new_data) - 1)
        self._data.extend(new_data)
        self.endInsertRows()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import logging
import os
import shutil
import sys
import time
from io import StringIO
from logging import error, info, warning
from typing import AbstractSet, Iterable, List, Optional, Tuple

from portmod.io import MergeIO, MergeMode, Message, Task, Transaction
from portmod.pybuild import Pybuild
from portmodlib.atom import Atom, QualifiedAtom, atom_sat
from portmodlib.colour import bright, green
from portmodlib.fs import onerror
from portmodlib.l10n import l10n
from portmodlib.module_util import _add_redirection

from ._deprecated.rebuild import get_rebuild_manifest
from ._deprecated.vfs import (
    clear_vfs_sort,
    require_vfs_sort,
    sort_vfs,
    vfs_needs_sorting,
)
from ._deps import resolve
from .cfg_protect import get_protected_path, get_redirections
from .config import get_config
from .config.license import _user_package_accept_license_path
from .config.profiles import get_system
from .config.sets import BUILTIN_SETS, add_set, get_set, remove_set
from .config.use import _user_package_use_path, add_use
from .download import (
    download_source,
    fetchable,
    find_download,
    find_download_and_fix_path,
    is_downloaded,
)
from .globals import env
from .loader import (
    AmbiguousAtom,
    _sandbox_execute_pybuild,
    load_all_installed,
    load_installed_pkg,
    load_pkg,
)
from .modules import (
    clear_module_updates,
    modules_need_updating,
    require_module_updates,
    update_modules,
)
from .package import install_pkg, remove_pkg
from .parsers.flags import add_flag
from .perms import Permissions
from .prompt import prompt_bool
from .query import get_flag_desc
from .repo.keywords import _user_package_accept_keywords_path, add_keyword
from .repo.loader import commit_moved
from .source import SourceManifest
from .transactions import (
    Delete,
    New,
    Transactions,
    UseDep,
    generate_transactions,
    print_transactions,
    sort_transactions,
)
from .tsort import CycleException
from .util import KeywordDep, LicenseDep, select_package
from .vdb import VDB


class InteractiveError(Exception):
    """
    Indicates that the merge cannot continue without user input

    Manual intervention is required
    """


def global_prepare():
    """Performs global system changes which must be done prior to changes"""
    commit_moved()


def global_updates():
    """Performs updates to global configuration"""
    # Fix vfs ordering and update modules

    try:
        sort_vfs()
        clear_vfs_sort()
    except CycleException as e:
        error(f"{e}")

    update_modules()
    clear_module_updates()

    redirections = get_redirections()
    if redirections:
        total_updates = sum(len(src) for src in redirections.values())
        warning(
            l10n(
                "pending-cfg-updates",
                total_updates=total_updates,
                total_files=len(redirections),
                command=f"portmod {env.PREFIX_NAME} cfg-update",
            )
        )


def deselect(pkgs: Iterable[str]):
    all_to_remove = []

    for name in pkgs:
        atom = Atom(name)
        to_remove = []
        for mod in get_set("selected-packages"):
            if atom_sat(mod, atom):
                to_remove.append(mod)

        if len(to_remove) == 1:
            info(">>> " + l10n("remove-from-world", atom=green(to_remove[0])))
            all_to_remove.append(to_remove[0])
        elif len(to_remove) > 1:
            raise AmbiguousAtom(atom, to_remove)

    if not all_to_remove:
        print(">>> " + l10n("no-matching-world-atom"))
        return

    if not env.INTERACTIVE or prompt_bool(bright(l10n("remove-from-world-qn"))):
        for mod in all_to_remove:
            remove_set("selected-packages", mod)


def _update_keywords(keyword_changes: Iterable[KeywordDep], protect: bool):
    keyword_file = _user_package_accept_keywords_path()
    protect_file = None
    if protect:
        protect_file = create_protect_file(keyword_file)

    for keyword in keyword_changes:
        add_keyword(keyword.atom, keyword.keyword, protect_file=protect_file)


def _update_licenses(
    license_changes: Iterable[LicenseDep],
    protect: bool,
):
    license_file = _user_package_accept_license_path()
    protect_file = None
    if protect:
        protect_file = create_protect_file(license_file)

    for license in license_changes:
        add_flag(protect_file or license_file, license.atom, license.license)


def _update_use_flags(use_changes: Iterable[UseDep], protect: bool):
    protect_file = None
    if protect:
        protect_file = create_protect_file(_user_package_use_path())

    for use in use_changes:
        add_use(
            use.flag.lstrip("-"),
            use.atom,
            use.flag.startswith("-"),
            use.comment,
            protect_file=protect_file,
        )


def create_protect_file(path: str):
    # get_protected_path will return the same path
    # if the original file does not exist
    if not os.path.exists(path):
        with open(path, "w", encoding="utf-8"):
            pass
    protect_file = get_protected_path(path)
    try:
        shutil.copy2(path, protect_file)
    except FileNotFoundError:
        pass
    _add_redirection(protect_file, path)
    return protect_file


def get_messages(pkg: Pybuild) -> List[Message]:
    """
    Produces messages produced by Pybuild1.warn or Pybuild1.info
    during package phase functions
    """
    warning_path = os.path.join(env.WARNINGS_DIR, pkg.ATOM.CPF)
    info_path = os.path.join(env.WARNINGS_DIR, pkg.ATOM.CPF)
    messages = []
    if os.path.exists(warning_path):
        with open(warning_path, "r") as file:
            messages.append(Message(Message.Type.WARNING, file.read()))

    if os.path.exists(info_path):
        with open(info_path, "r") as file:
            messages.append(Message(Message.Type.INFO, file.read()))
    return messages


def download_files(
    io: MergeIO,
    sources: List[Tuple[Transaction, SourceManifest]],
) -> List[Tuple[Transaction, SourceManifest]]:
    """Downloads the provided files and returns the ones which failed"""
    info("Downloading files...")
    failed = []
    for index, (trans, source) in enumerate(
        reversed(sorted(sources, key=lambda x: x[1].size))
    ):
        try:
            download_source(
                trans.pkg,
                source,
                get_progress=io.get_stepped_fetch_progress(index + 1, len(sources)),
            )
        except Exception as e:
            error(e)
            failed.append((trans, source))
    return failed


def _print_nofetch(
    io: MergeIO, transactions: Iterable[Transaction], skip_fetchable: bool = True
):
    for trans in transactions:
        if (
            not isinstance(trans, Delete)
            and "pkg_pretend" in trans.pkg.FUNCTIONS
            and (not skip_fetchable or fetchable(trans.pkg, trans.flags))
        ):
            # TODO: There are various variables that should be set on mod during pkg_pretend
            info(">>> " + l10n("pkg-pretend", atom=green(trans.pkg.ATOM.CPF)))
            proc = _sandbox_execute_pybuild(
                trans.pkg.FILE,
                "pretend",
                Permissions(),
                save_state=True,
                init={"USE": trans.flags},
                pipe_output=True,
            )
            io.pkg_pretend(trans.pkg.ATOM, proc.read_output(timeout=5))
            proc.wait()


def _deprecated_rebuild(io: MergeIO, merged: List[Transaction]):
    # TODO: This should also be skipped if there are no Pybuild1 packages, but it's
    # hiding an issue with the vfs-rebuild module
    # Check if packages were just modified and can be removed from the rebuild set
    # Any transaction type warrants removal, as they were either rebuilt,
    # and thus can be removed, or deleted, and no longer need to be rebuilt
    for atom in get_set("rebuild"):
        installed_pkg = load_installed_pkg(atom)
        if (
            not installed_pkg
            or installed_pkg.CPN in [trans.pkg.CPN for trans in merged]
            and installed_pkg._PYBUILD_VER == 1
        ):
            remove_set("rebuild", atom)

    if any(pkg._PYBUILD_VER == 1 for pkg in load_all_installed()):
        info(l10n("checking-rebuild"))
    else:
        # If no pybuild1 packages are installed, the rebuild feature isn't needed
        return

    # Check if packages need to be added to rebuild set
    for pkg in load_all_installed():
        if pkg.CPN not in get_set("rebuild") and pkg.INSTALLED_REBUILD_FILES:
            seen = set()
            for entry in get_rebuild_manifest(
                pkg.get_installed_env().get("REBUILD_FILES", [])
            ):
                seen.add(entry.name)
                if pkg.INSTALLED_REBUILD_FILES.get(entry.name) != entry:
                    add_set("rebuild", pkg.CPN)
                    break
            if not all(entry in seen for entry in pkg.INSTALLED_REBUILD_FILES.entries):
                add_set("rebuild", pkg.CPN)

    if get_set("rebuild"):
        io.rebuild_warning(sorted(get_set("rebuild")), l10n("rebuild-message"))


def _expand_sets(atoms: Iterable[str]) -> List[Atom]:
    targetlist = []
    for modstr in atoms:
        if modstr.startswith("@"):
            # Atom is actually a set. Load set instead
            set_contents = get_set(modstr[1:])
            targetlist.extend(_expand_sets(set_contents))
        else:
            targetlist.append(Atom(modstr))
    return targetlist


def _resolve_nodeps(
    to_install: List[Atom],
    selected: AbstractSet[Atom],
    *,
    oneshot: bool,
    depclean: bool,
    update: bool,
    emptytree: bool,
) -> Transactions:
    fqatoms = []
    enabled_flags = {}
    usedeps = []
    for atom in to_install:
        pkg, _ = select_package(load_pkg(atom))
        fqatoms.append(pkg.ATOM)
        enabled_flags[pkg.ATOM] = pkg.get_use()
        for flag in atom.USE:
            usedeps.append(
                UseDep(
                    pkg.ATOM,
                    flag,
                    description=get_flag_desc(pkg, flag),
                    oldvalue=None,
                    comment=("# Use requirement passed on command line",),
                )
            )
            if flag.startswith("-"):
                enabled_flags[pkg.ATOM].remove(flag.lstrip("-"))
            else:
                enabled_flags[pkg.ATOM].add(flag)

    selected_cpn = set()
    for atom in selected:
        pkg, _ = select_package(load_pkg(atom))
        selected_cpn.add(QualifiedAtom(pkg.CPN))

    return generate_transactions(
        fqatoms,
        [],
        set() if oneshot or depclean else selected_cpn,
        usedeps,
        enabled_flags,
        update=update,
        emptytree=emptytree,
    )


def _get_restricted_fetch(
    io: MergeIO, transactions: Transactions, include_fetchable: bool = False
):
    # Check for restricted fetch packages and print their nofetch notices
    restricted = 0
    for trans in transactions.pkgs:
        if not isinstance(trans, Delete) and "pkg_nofetch" in trans.pkg.FUNCTIONS:
            can_fetch = fetchable(trans.pkg, trans.flags)
            sources = trans.pkg.get_source_manifests(trans.flags)
            to_fetch = [
                source
                for source in sources
                if find_download_and_fix_path(source) is None
            ]
            if include_fetchable:
                restricted_sources = set(to_fetch)
            else:
                restricted_sources = set(to_fetch) - set(can_fetch)
            if restricted_sources and not is_downloaded(trans.pkg, trans.flags):
                restricted += len(restricted_sources)
                state = {
                    "UNFETCHED": to_fetch,
                    "A": sources,
                    "USE": trans.flags,
                }
                proc = _sandbox_execute_pybuild(
                    trans.pkg.FILE,
                    "nofetch",
                    Permissions(),
                    init=state,
                    pipe_output=True,
                )
                io.pkg_nofetch(trans.pkg.ATOM, proc.read_output(timeout=5))
                proc.wait()
    return restricted


def _check_downloads(
    io: MergeIO,
    transactions: Transactions,
    failed_downloads: List[Tuple[Transaction, SourceManifest]],
    restricted_downloads: int,
):
    # Any downloads which failed should be eligible for a restricted fetch message, if available
    # For any other failed downloads, print their URL and give the user the opportunity to manually fetch them
    if failed_downloads and not env.INTERACTIVE:
        raise InteractiveError(l10n("exiting-non-interactive-nofetch"))

    _print_nofetch(io, transactions.pkgs)
    for trans, source in failed_downloads:
        if "pkg_pretend" in trans.pkg.FUNCTIONS:
            error(
                f"File {source.name} from {trans.pkg} failed to download, but manual fetch instructions were provided:"
            )
            _print_nofetch(io, [trans], skip_fetchable=False)
        else:
            error(
                f"File {source.name} failed to download from {source.url}. If you are able to do so manually"
            )

    if restricted_downloads > 0:
        warning(l10n("restricted-fetch-summary", restricted=restricted_downloads))
        # Update restricted download count.
        new_restricted_downloads = _get_restricted_fetch(
            io, transactions, include_fetchable=True
        )
        if new_restricted_downloads >= restricted_downloads:
            raise InteractiveError(l10n("restricted-fetch-unchanged"))
        return new_restricted_downloads
    return 0


def merge(
    atoms: Iterable[str],
    *,
    io: MergeIO,
    delete: bool = False,
    depclean: bool = False,
    auto_depclean: bool = False,
    oneshot: bool = False,
    verbose: bool = False,
    update: bool = False,
    nodeps: bool = False,
    deselect: Optional[bool] = None,
    select: Optional[bool] = None,
    deep: bool = False,
    emptytree: bool = False,
):
    global_prepare()

    # Ensure that we always get the config before performing operations on packages
    # This way the config settings will be available as environment variables.
    get_config()

    atomlist = _expand_sets(atoms)
    explicit = set(atomlist)
    selected = set(Atom(atom) for atom in atoms if not atom.startswith("@"))
    selected_sets = set(atom[1:] for atom in atoms if atom.startswith("@"))

    to_remove = set()
    if delete or depclean:
        for atom in atomlist:
            skip = False
            for system_atom in get_system():
                if atom_sat(system_atom, atom):
                    warning(l10n("skipping-system-package", atom=system_atom))
                    skip = True
                    break

            if not skip:
                to_remove.add(atom)
        to_install = []
    else:
        to_install = atomlist

    if delete:
        # Do nothing. We don't care about deps
        transactions = Transactions()
        for atom in to_remove:
            ipkg = load_installed_pkg(atom)
            if not ipkg:
                raise Exception(l10n("not-installed", atom=atom))
            transactions.append(Delete(ipkg))
    elif nodeps:
        transactions = _resolve_nodeps(
            to_install,
            selected,
            oneshot=oneshot,
            depclean=depclean,
            emptytree=emptytree,
            update=update,
        )
    else:
        transactions = resolve(
            to_install,
            to_remove,
            explicit,
            set() if oneshot or depclean else selected,
            set() if oneshot or depclean else selected_sets,
            deep=deep
            or (depclean and not to_remove),  # No argument depclean implies deep
            update=update,
            depclean=auto_depclean or depclean,
            emptytree=emptytree,
        )

    transactions = sort_transactions(transactions)

    # Inform user of changes
    if transactions.pkgs:
        # Don't print transaction list when in quiet mode and no-confirm is passed
        if env.INTERACTIVE or logging.root.level < logging.WARN:
            if delete or depclean:
                io.display_transactions(
                    MergeMode.REMOVE, transactions.pkgs, transactions.new_selected
                )
            else:
                io.display_transactions(
                    MergeMode.INSTALL, transactions.pkgs, transactions.new_selected
                )
    elif (vfs_needs_sorting() or modules_need_updating()) and not transactions.pkgs:
        global_updates()
        io.finished(l10n("nothing-else-to-do"))
        return
    elif not transactions.pkgs:
        io.finished(l10n("nothing-to-do"))
        return

    tasks = []

    if transactions.config:
        protect = not env.INTERACTIVE and not env.TESTING

        keyword_changes = list(
            filter(lambda x: isinstance(x, KeywordDep), transactions.config)
        )
        if keyword_changes:
            keyword_task = Task(_update_keywords, keyword_changes, protect)
            tasks.append(keyword_task)
            io.keyword_changes(keyword_changes, keyword_task)

        license_changes = list(
            filter(lambda x: isinstance(x, LicenseDep), transactions.config)
        )
        if license_changes:
            license_task = Task(_update_licenses, license_changes, protect)
            tasks.append(license_task)
            io.license_changes(license_changes, license_task)

        use_changes = list(filter(lambda x: isinstance(x, UseDep), transactions.config))
        if use_changes:
            use_task = Task(_update_use_flags, use_changes, protect)
            tasks.append(use_task)
            io.use_changes(use_changes, use_task)

        # Note: Only CLI should be able to be non-interactive.
        if protect:
            # In non-interactive mode, don't continue if configuration changes are required
            while not all(task.done for task in tasks):
                time.sleep(0.01)
            info("")
            warning(
                l10n(
                    "configuration-changes-required",
                    command=f"portmod {env.PREFIX_NAME} cfg-update",
                ),
            )
            sys.exit(1)

    # Create a temporary VDB so that common packages needed for pkg_nofetch and pkg_pretend
    # but that haven't yet been installed can be found by the loader
    info("Making new common packages available to the packaging environment...")
    if os.path.exists(env.TMP_VDB):
        shutil.rmtree(env.TMP_VDB, onerror=onerror)

    for trans in transactions.pkgs:
        if isinstance(trans, New) and trans.pkg.CATEGORY == "common":
            os.makedirs(os.path.join(env.TMP_VDB, "common", trans.pkg.PN))
            shutil.copy2(
                trans.pkg.FILE, os.path.join(env.TMP_VDB, "common", trans.pkg.PN)
            )

    restricted_downloads = _get_restricted_fetch(io, transactions)

    info("Checking for restricted-fetch downloads...")
    tmp_dir = env.TMP_DIR
    # If TMP_DIR doesn't exist, either use the parent, or if that doesn't exist,
    # just create it
    if not os.path.exists(env.TMP_DIR):
        if os.path.exists(os.path.dirname(env.TMP_DIR)):
            tmp_dir = os.path.dirname(env.TMP_DIR)
        else:
            os.makedirs(tmp_dir, exist_ok=True)
    tmp_space = shutil.disk_usage(tmp_dir).free

    # TODO: Skip packages which can be downloaded automatically but also have a nofetch message
    # If the download fails, then print the nofetch message before entering the feedback loop
    pre_merge_warnings = 0

    for transaction in transactions.pkgs:
        total_size = 0
        for source in transaction.pkg.get_source_manifests(transaction.flags):
            total_size += source.size

        # We assume that files have a compression ratio of approximately 0.5
        # Thus we want at least twice the size of the archives in free space.
        if total_size * 2 > tmp_space:
            pre_merge_warnings += 1
            io.space_warning(
                transaction.pkg.ATOM,
                l10n(
                    "tmp-space-too-small",
                    dir=env.TMP_DIR,
                    free=tmp_space / 1024 / 1024,
                    size=total_size * 2 / 1024 / 1024,
                ),
            )
        pre_merge_warnings += len(
            {
                message
                for message in get_messages(transaction.pkg)
                if message.typ == Message.Type.WARNING
            }
        )

    if pre_merge_warnings:
        warning(l10n("pre-merge-warnings", warnings=pre_merge_warnings))

    if restricted_downloads and not env.INTERACTIVE:
        _print_nofetch(io, transactions.pkgs)
        raise InteractiveError(l10n("exiting-non-interactive-nofetch"))

    # Wait for initial tasks to complete
    while not all(task.done for task in tasks):
        time.sleep(0.01)
    tasks.clear()

    to_download = []
    failed_downloads: List[Tuple[Transaction, SourceManifest]] = []
    for trans in transactions.pkgs:
        if not isinstance(trans, Delete):
            for source in fetchable(trans.pkg, trans.flags):
                if find_download(source) is None:
                    to_download.append((trans, source))

    if to_download:
        download_task = Task(download_files, io=io, sources=to_download)
        io.download_ready(download_task)
        failed_downloads = download_task.result()

    while env.INTERACTIVE:
        ready_task = Task(
            _check_downloads,
            io,
            transactions,
            failed_downloads=failed_downloads,
            restricted_downloads=restricted_downloads,
        )
        # Only print failed downloads the first time
        failed_downloads = []
        io.merge_ready(ready_task)
        while not ready_task.done:
            time.sleep(0.01)
        restricted_downloads = ready_task.result()
        if restricted_downloads <= 0:
            break

    err = None
    merged: List[Transaction] = []
    try:
        # Install (or remove) packages in order
        for trans in transactions.pkgs:
            if isinstance(trans, Delete):
                remove_pkg(trans.pkg, io=io.get_remove_io(trans.pkg.ATOM))
                if deselect is None or deselect:
                    if trans.pkg.CPN in get_set("selected-packages"):
                        info(
                            ">>> "
                            + l10n("remove-from-world", atom=green(trans.pkg.CPN))
                        )
                        remove_set("selected-packages", trans.pkg.CPN)
                merged.append(trans)
            else:
                install_pkg(
                    trans.pkg, trans.flags, io=io.get_install_io(trans.pkg.ATOM)
                )

                if trans.pkg in transactions.new_selected and not oneshot:
                    if trans.pkg.CPN not in get_set("selected-packages"):
                        info(">>> " + l10n("add-to-world", atom=green(trans.pkg.CPN)))
                        add_set("selected-packages", trans.pkg.CPN)
                merged.append(trans)

            require_module_updates()
            if trans.pkg._PYBUILD_VER == 1:
                require_vfs_sort()
                sort_vfs()
                clear_vfs_sort()
    # Unable to install package. Aborting installing remaining packages
    except Exception as e:
        err = trans.pkg.ATOM
        raise e
    # Run final code even if package installation fails
    finally:
        for trans in merged:
            messages = get_messages(trans.pkg)

            if messages:
                io.pkg_messages(trans.pkg.ATOM, messages)

        for set_name in selected_sets:
            if set_name not in BUILTIN_SETS:
                if deselect or delete or depclean:
                    remove_set("selected-sets", Atom(set_name))
                else:
                    add_set("selected-sets", Atom(set_name))

        # If no packages successfully merged, there's nothing to update in the VDB
        with StringIO() as transstring:
            print_transactions(
                merged, set(), verbose=True, out=transstring, summarize=False
            )
            if err:
                # There was an error. We report the packages that were successfully merged and
                # note that an error occurred, however we still commit anyway.
                summary = l10n("merge-success-and-error", num=len(merged), atom=err)
            else:
                summary = l10n("merge-success", num=len(merged))

            if merged:
                # Commit changes to installed database
                with VDB(summary + "\n" + transstring.getvalue()):
                    pass
            if err:
                info(summary)
                io.finished(summary)

    _deprecated_rebuild(io, merged)
    global_updates()
    info(summary)
    io.finished(summary)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Portmod config tests
"""

import os
import sys
from logging import error

import pytest

from portmod._cli.inquisitor import main, scan_file, scan_path, scan_sources

from .env import TEST_REPO, connected, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_inquisitor():
    """
    Basic inquisitor test on test repo
    """
    has_error = False

    def err(string: str):
        nonlocal has_error
        error(string)
        has_error = True

    scan_path(TEST_REPO.location, TEST_REPO.location, err)
    for path_root, _, filenames in os.walk(TEST_REPO.location):
        scan_path(path_root, TEST_REPO.location, err)
        os.chdir(path_root)
        for filename in filenames:
            scan_file(filename, TEST_REPO.location, err)
            scan_path(filename, TEST_REPO.location, err)
    if has_error:
        raise Exception("Inqusitor failed. See error log for details.")


def test_inquisitor_main():
    """
    Basic inquisitor test on test repo
    """
    sys.argv = ["inquisitor", "scan", TEST_REPO.location]
    main()


@pytest.mark.skipif(not connected(), reason="Requires network access")
def test_scan_sources():
    has_error = False

    def err(string: str):
        nonlocal has_error
        error(string)
        has_error = True

    scan_sources(
        os.path.join(
            TEST_REPO.location,
            "test",
            "quill-of-feyfolken",
            "quill-of-feyfolken-2.0.2.pybuild",
        ),
        err,
    )
    assert not has_error, "Source should be available"


# TODO: Add tests that modify the test repo to be invalid



from typing import Any, List, Union

from PySide6.QtCore import (
    QAbstractListModel,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
    Slot,
)

from portmod._gui.packages import search_packages
from portmodlib.atom import QualifiedAtom
from portmodlib.portmod import PackageIndexData


class SearchedPackagesModel(QAbstractListModel):
    def __init__(self) -> None:
        super(SearchedPackagesModel, self).__init__()
        self._data: List[PackageIndexData] = []

    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: Qt.ItemDataRole) -> Any:  # type: ignore
        if role == Qt.ItemDataRole.DisplayRole:
            index_data = self._data[index.row()]
            value = f"{index_data.name}\n\n{index_data.desc}"
            if isinstance(value, QualifiedAtom):
                return str(value)

            return value

    def rowCount(self, index: Union[QModelIndex, QPersistentModelIndex]) -> int:  # type: ignore
        return len(self._data)

    @Slot(str)
    def updateSearchTerm(self, searchTerm: str) -> None:
        self.beginRemoveRows(QModelIndex(), 0, len(self._data))
        self._data.clear()
        self.endRemoveRows()

        new_data = search_packages(searchTerm)

        self.beginInsertRows(QModelIndex(), 0, len(new_data) - 1)
        self._data.extend(new_data)
        self.endInsertRows()

    @Slot(int)
    def getAtom(self, rowIndex: int) -> str:
        index = self.index(rowIndex)
        # The atom is stored in the first index of the list
        atom: str = self._data[index.row()].cpn
        return atom



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for directly loading pybuild files.
These functions should not be called directly.
See portmod.loader for functions to load pybuilds safely using a sandbox.
"""

import ast
import glob
import os
import sys
from functools import lru_cache
from logging import error, warning
from types import SimpleNamespace, new_class
from typing import Any, Dict, Generator, List, Optional, Union, cast

from RestrictedPython import (
    RestrictingNodeTransformer,
    compile_restricted_exec,
    limited_builtins,
    safe_globals,
)
from RestrictedPython.Eval import default_guarded_getitem, default_guarded_getiter
from RestrictedPython.Guards import (
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence,
    safer_getattr,
)

from .atom import Atom
from .globals import tmp_vdb, vdb_path
from .portmod import parse_yaml_mapping
from .pybuild import FullInstalledPybuild, FullPybuild


def _iterate_pybuilds(
    directory: str, *, recursive: bool = False
) -> Generator[str, None, None]:
    """
    Iterates over packages by file

    Fast and imprecise. You should probably use iterate_pybuilds or at least
    call is_package on the results
    Unlike iterate_pybuilds, will ignore things such as moved packages
    args:
        directory: A glob-like path as can be consumed by the glob package.
                   All directories matching this glob will be scanned
    """
    for directory in glob.iglob(directory, recursive=recursive):
        if not os.path.isdir(directory):
            continue
        basename = os.path.basename(directory.rstrip(os.sep + (os.altsep or "")))
        for path in os.listdir(directory):
            # Only match paths which begin with the package directory
            # and end with .pybuild or .yaml
            if path.startswith(basename):
                name, ext = os.path.splitext(path)
                if ext in (".pybuild", ".yaml"):
                    yield os.path.join(directory, path)


def find_installed_path(atom: Atom, vdb: Optional[str] = None) -> Optional[str]:
    if not vdb:
        vdb = vdb_path()

    results = list(_iterate_pybuilds(os.path.join(vdb, atom.C or "*", atom.PN)))
    if results:
        if len(results) > 1:
            error(
                f"Multiple package files matched for the installed package {atom}. Ignoring."
            )
            return None
        return results[0]
    return None


@lru_cache()
def _import_common(name: str) -> SimpleNamespace:
    """
    args:
        name: The import name as an absolute import path
        installed: Whether or not the package calling this is an installed package
        load_function: The function taking a file path and a keyword argument installed,
            indicating the installed status of the file to be loaded, to be used to load
            the common module

    returns:
        The Module as a SimpleNamespace
    """
    if len(name.split(".")) > 2:
        raise ImportError(f"Invalid package {name}")
    _, module_name = name.split(".")
    base_atom = Atom(f"common/{module_name}")
    path = find_installed_path(base_atom)

    if not path:
        path = find_installed_path(base_atom, tmp_vdb())

    if not path:
        raise ImportError(f"Could not find package {name}")

    result = __load_module(path)

    return SimpleNamespace(
        **{key: value for key, value in result.items() if not key.startswith("_")}
    )


def _import(repo: Optional[str] = None):
    def import_fn(name, globs=None, loc=None, fromlist=(), level=0):
        if name.startswith("common."):
            return _import_common(name)
        return __import__(name, globs, loc, fromlist, level)

    return import_fn


# Default implementation to handle invalid pybuilds
class Package:
    def __init__(self):
        raise Exception("Package is not defined")


def default_write_guard(ob):
    """Noop write guard"""
    return ob


def safer_hasattr(obj, name):
    """
    Version of hasattr implemented using safet_getattr

    This doesn't really provide any extra security, but does mean that
    str.format, and attributes beginning with underscores, which are
    blocked by safer_getattr, return False rather than True
    """
    try:
        safer_getattr(obj, name)
    except (NotImplementedError, AttributeError):
        return False
    return True


def safer_dir(obj):
    """
    Version of dir which doesn't report underscored attributes
    """
    return [str(elem) for elem in dir(obj) if not elem.startswith("_")]


def default_apply(func, *args, **kwargs):
    return func(*args, **kwargs)


MINIMAL_GLOBALS: Dict[str, Any] = safe_globals
SAFE_GLOBALS: Dict[str, Any] = MINIMAL_GLOBALS.copy()
SAFE_GLOBALS["__builtins__"] = MINIMAL_GLOBALS["__builtins__"].copy()
SAFE_GLOBALS.update({"Package": Package})
SAFE_GLOBALS["__builtins__"].update(
    {
        "FileNotFoundError": FileNotFoundError,
        "__metaclass__": type,
        "_apply_": default_apply,
        "_getattr_": safer_getattr,
        "_getitem_": default_guarded_getitem,
        "_getiter_": default_guarded_getiter,
        "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
        "_unpack_sequence_": guarded_unpack_sequence,
        "_write_": default_write_guard,
        "all": all,
        "any": any,
        "dict": dict,
        "enumerate": enumerate,
        "filter": filter,
        "frozenset": frozenset,
        "getattr": safer_getattr,
        "hasattr": safer_hasattr,
        "iter": iter,
        "map": map,
        "max": max,
        "min": min,
        "next": next,
        "reversed": reversed,
        "set": set,
        "sorted": sorted,
        "sum": sum,
        "super": super,
        "dir": safer_dir,
    }
)
SAFE_GLOBALS["__builtins__"].update(limited_builtins)

SANDBOX_GLOBALS: Dict[str, Any] = SAFE_GLOBALS.copy()
SANDBOX_GLOBALS["__builtins__"] = SAFE_GLOBALS["__builtins__"].copy()


class PrintWrapper:
    def __init__(self, _getattr_=None):
        self.txt = []
        self._getattr_ = _getattr_

    def write(self, text):
        self.txt.append(text)

    def __call__(self):
        return "".join(self.txt)

    def _call_print(self, *objects, **kwargs):
        if kwargs.get("file", None) is None:
            kwargs["file"] = sys.stdout
        else:
            self._getattr_(kwargs["file"], "write")
        print(*objects, **kwargs)


# print and open are only allowed within the sandbox
SANDBOX_GLOBALS["__builtins__"].update({"_print_": PrintWrapper, "open": open})


class Policy(RestrictingNodeTransformer):
    def visit_JoinedStr(self, node):
        return self.node_contents_visit(node)

    def visit_FormattedValue(self, node):
        return self.node_contents_visit(node)

    def visit_AnnAssign(self, node):
        return self.node_contents_visit(node)

    def visit_AugAssign(self, node):
        return self.node_contents_visit(node)

    def visit_FunctionDef(self, node):
        node = RestrictingNodeTransformer.visit_FunctionDef(self, node)
        if node.name == "__init__":
            newnode = ast.parse("super().__init__()").body[0]
            # The interpreter needs correct line and column numbers
            # for displaying traces.
            # The inserted code doesn't actually exist in the source, but setting it
            # to the previous line's position is close enough.
            newnode.lineno = node.lineno
            newnode.col_offset = node.col_offset
            newnode.end_lineno = node.end_lineno
            newnode.end_col_offset = node.end_col_offset
            node.body.insert(0, newnode)
        return node


def restricted_load(code, filepath: str, _globals: Dict[str, Any]):
    byte_code, errors, warnings, names = compile_restricted_exec(
        code, filename=filepath, policy=Policy
    )
    if errors:
        raise SyntaxError(errors)
    seen: Dict[str, str] = {}
    for message in [seen.setdefault(x, x) for x in warnings if x not in seen]:
        if not message.endswith("Prints, but never reads 'printed' variable."):
            warning(f"In file {filepath}: {message}")
    exec(byte_code, _globals, _globals)  # nosec B102


def __load_module(path: str) -> Dict[str, Any]:
    filename, _ = os.path.splitext(os.path.basename(path))

    with open(path, "r", encoding="utf-8") as file:
        code = file.read()
        tmp_globals = SANDBOX_GLOBALS.copy()
        tmp_globals["__builtins__"]["__import__"] = _import()
        tmp_globals["__name__"] = filename
        restricted_load(code, path, tmp_globals)

    return tmp_globals


def __load_yaml_module(path: str, _import) -> Dict[str, Any]:
    from string import Template

    from pybuild.info import _get_info

    filename, _ = os.path.splitext(os.path.basename(path))

    atom = Atom(os.path.basename(filename))

    with open(path) as file:
        # Use string templates to fill in stuff from pybuild.info
        file_data = Template(file.read()).safe_substitute(_get_info(atom))
        yaml_data = parse_yaml_mapping(file_data, path)

    # Continue substituting  until either all placeholders are gone,
    # or until the number of placeholders doesn't reduce after a substitution.
    # Limit 10 substitutions in case the placeholders only increase.
    placeholder_count = file_data.count("$")
    substitutions = 1
    while placeholder_count > 0 and substitutions < 10:
        # Subsitute top-level string values
        substitutions += 1
        values: Dict[str, Union[str, int, float]] = {
            key: value
            for key, value in yaml_data.items()
            if isinstance(value, (str, int, float))
        }

        file_data = Template(file_data).safe_substitute(values)
        yaml_data = parse_yaml_mapping(file_data, path)
        new_placeholder_count = file_data.count("$")
        if placeholder_count == new_placeholder_count:
            break

    try:
        superclass_data = yaml_data["inherit"]
    except KeyError:
        raise ValueError("Yaml packages must inherit from a package class!")

    if isinstance(superclass_data, str):
        superclasses: List[str] = [superclass_data]
    else:
        superclasses = superclass_data

    imports = []
    superclass_objects = []
    for superclass in superclasses:
        package, _, class_name = superclass.rpartition(":")
        module = _import(package, fromlist=[class_name])
        superclass_objects.append(getattr(module, class_name))
        imports.append(package)

    return {
        "Package": new_class(
            "Package",
            tuple(superclass_objects),
            exec_body=lambda ns: ns.update(yaml_data),
        ),
        # Used for generating common dependencies later
        "imports": imports,
    }


def __load_file_common(path: str, module: Dict[str, Any], installed: bool):
    module["Package"].__pybuild__ = path
    pkg = module["Package"]()
    pkg.FILE = os.path.abspath(path)
    pkg.INSTALLED = False

    if not installed:
        # determine common dependencies
        def find_common_imports(file: str):
            depends = []
            _, ext = os.path.splitext(path)
            if ext == ".pybuild":
                with open(file, "r", encoding="utf-8") as fp:
                    tree = ast.parse(fp.read())

                def find_imports(tree: ast.AST):
                    if isinstance(tree, ast.Module):
                        for statement in tree.body:
                            if isinstance(statement, ast.Import):
                                for alias in statement.names:
                                    if alias.name.startswith("common."):
                                        depends.append(alias.name.replace(".", "/"))
                            elif isinstance(statement, ast.ImportFrom):
                                if statement.module and statement.module.startswith(
                                    "common."
                                ):
                                    depends.append(statement.module.replace(".", "/"))

                # Globals
                find_imports(tree)
                # TODO: Inline imports in functions?
            else:
                return [
                    module.replace(".", "/")
                    for module in module["imports"]
                    if module.startswith("common.")
                ]
            return depends

        pkg.DEPEND = " ".join([pkg.DEPEND] + find_common_imports(pkg.FILE))
    return pkg


def load_file(path: str, *, installed=False) -> FullPybuild:
    """
    Loads a pybuild file

    :param path: Path of the pybuild file
    """
    _, ext = os.path.splitext(path)
    if ext == ".pybuild":
        module = __load_module(path)
    elif ext == ".yaml":
        module = __load_yaml_module(path, _import=_import())
    else:
        raise NotImplementedError(f"Cannot load file {path} with type {ext}")
    pkg = __load_file_common(path, module, installed)
    return cast(FullPybuild, pkg)


def load_installed(file: str) -> FullInstalledPybuild:
    """
    Loads an installed pybuild

    :param file: Path of the pybuild file
    """
    mod = cast(FullInstalledPybuild, load_file(file, installed=True))
    __load_installed_common(mod, file)
    return mod


def __load_installed_common(mod, file: str):
    """Shared code for loading installed pybuilds"""
    mod.INSTALLED = True
    parent = os.path.dirname(file)

    def read_file(name: str) -> Optional[str]:
        if os.path.exists(os.path.join(parent, name)):
            with open(os.path.join(parent, name), "r") as repo_file:
                return repo_file.read().strip()
        return None

    repo = read_file("REPO")
    if not repo:
        raise Exception(
            f"Internal Error: Installed package in file {file}"
            "has no repository identifier"
        )
    mod.REPO = repo
    mod.INSTALLED_USE = set((read_file("USE") or "").split())
    mod.RDEPEND = read_file("RDEPEND") or mod.RDEPEND
    mod.DEPEND = read_file("DEPEND") or mod.DEPEND



from typing import Any, Dict, Tuple, Union

from PySide6.QtCore import (
    QAbstractListModel,
    QByteArray,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
)


class FlagListModel(QAbstractListModel):
    def __init__(self, use_flags: Dict[str, Tuple[bool, str]]):
        super().__init__()

        self._data = use_flags

    def rowCount(self, parent: Union[QModelIndex, QPersistentModelIndex]) -> int:  # type: ignore
        return len(self._data)

    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: int) -> Any:  # type: ignore
        if role == Qt.ItemDataRole.DisplayRole:
            return list(self._data.keys())[index.row()]
        elif role == Qt.ItemDataRole.CheckStateRole:
            state = list(self._data.values())[index.row()][0]
            if state:
                return Qt.CheckState.Checked
            else:
                return Qt.CheckState.Unchecked
        elif role == Qt.ItemDataRole.UserRole:
            return list(self._data.values())[index.row()][1]

    def roleNames(self) -> Dict[int, QByteArray]:
        result: Dict[int, Any] = super().roleNames()
        result[Qt.ItemDataRole.CheckStateRole] = QByteArray(b"checkState")
        result[Qt.ItemDataRole.UserRole] = QByteArray(b"description")
        return result



# Copyright 2019-2023 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Match, Optional, Tuple

ext_ver_re = (
    r"(?P<NUMERIC>(\d+)(\.\d+)*)"
    r"(?P<LETTER>[a-z])?"
    # Note: The p suffix is disabled for external versions
    r"(?P<SUFFIX>((_(pre|beta|alpha|rc)\d*)*))"
)
ver_re = (
    r"(e(?P<EPOCH>\d+)-)?"
    r"(?P<NUMERIC>(\d+)(\.\d+)*)"
    r"(?P<LETTER>[a-z])?"
    r"(?P<SUFFIX>((_(pre|p|beta|alpha|rc)\d*)*))"
    r"(-r(?P<REV>\d+))?"
)
_ver_re = re.compile("^" + ver_re + "$")
_ext_ver_re = re.compile("^" + ext_ver_re + "$")


@dataclass
class Version:
    """
    A package version

    This class should be treated as immutable
    """

    numeric: List[str]
    suffixes: List[str]
    epoch: Optional[int] = None
    letter: str = ""
    revision: Optional[int] = None

    def __init__(
        self,
        version: str,
        *,
        _match: Optional[Match[str]] = None,
        external: bool = False,
    ):
        """
        args:
            version: A string representation of the version to be parsed
            external: If true, epochs and revisions  will be disabled, for use with upstream versions.
        """
        if not _match:
            if external:
                match = _ext_ver_re.match(version)
            else:
                match = _ver_re.match(version)
            if match is None:
                raise ValueError(f"Version {version} could not be parsed")
        else:
            match = _match

        if not external and match.group("EPOCH"):
            self.epoch = int(match.group("EPOCH"))
        self.numeric = list(str(match.group("NUMERIC")).split("."))
        self.letter = match.group("LETTER") or ""
        suffixes = match.group("SUFFIX")
        if suffixes:
            self.suffixes = suffixes.lstrip("_").split("_")
        else:
            self.suffixes = []
        if not external and match.group("REV"):
            self.revision = int(match.group("REV"))

    def __hash__(self):
        return hash(
            (
                self._normalize_numeric(),
                self.suffixes,
                self.epoch,
                self.letter,
                self.revision,
            )
        )

    def __str__(self) -> str:
        return self.display()

    def display(self, *, revision: bool = True, epoch: bool = True) -> str:
        return (
            (f"e{self.epoch}-" if self.epoch is not None and epoch else "")
            + ".".join(self.numeric)
            + (self.letter or "")
            + ("_" + "_".join(self.suffixes) if self.suffixes else "")
            + (f"-r{self.revision}" if self.revision is not None and revision else "")
        )

    def _normalize_numeric(self) -> List[str]:
        return list(map(lambda x: x.rstrip("0"), self.numeric))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return self.compare(other)
        return False

    def compare(self, other: "Version", *, ignore_revision=False) -> bool:
        # Note: versions with different numbers of numeric components are not equivalent
        # according to the PMS, even if the extra components are trailing 0s.
        return (
            self.epoch == other.epoch
            and self._normalize_numeric() == other._normalize_numeric()
            and self.letter == other.letter
            and self.suffixes == other.suffixes
            and (self.revision == other.revision or ignore_revision)
        )

    def __ge__(self, other: "Version") -> bool:
        return self > other or self == other

    def __gt__(self, other: "Version") -> bool:
        return self.greater(other)

    def greater(self, other: "Version", ignore_extra_components=False) -> bool:
        # Compare epochs
        if (self.epoch or 0) > (other.epoch or 0):
            return True
        if (self.epoch or 0) < (other.epoch or 0):
            return False

        if int(self.numeric[0]) > int(other.numeric[0]):
            return True
        if int(self.numeric[0]) < int(other.numeric[0]):
            return False

        # Compare numeric components
        for index, val in enumerate(self.numeric[1:], start=1):
            if index >= len(other.numeric):
                if ignore_extra_components:
                    return False
                else:
                    return True
            if val.startswith("0") or other.numeric[index].startswith("0"):
                # If either starts with a leading 0, strip trailing zeroes and
                # compare lexicographically.
                # I.e. 01 should always come before 10
                my_stripped = val.rstrip("0")
                other_stripped = other.numeric[index].rstrip("0")
                if my_stripped > other_stripped:
                    return True
                if my_stripped < other_stripped:
                    return False
            else:
                if int(val) > int(other.numeric[index]):
                    return True
                if int(val) < int(other.numeric[index]):
                    return False
        if len(other.numeric) > len(self.numeric):
            return False

        if ignore_extra_components:
            # Any further
            return False

        # Compare letter components
        if self.letter > other.letter:
            return True
        if self.letter < other.letter:
            return False

        # Compare suffixes
        for a_s, b_s in zip(self.suffixes, other.suffixes):
            asm = re.match(r"(?P<S>[a-z]+)(?P<N>\d+)?", a_s)
            bsm = re.match(r"(?P<S>[a-z]+)(?P<N>\d+)?", b_s)
            assert asm
            assert bsm
            a_suffix = asm.group("S")
            b_suffix = bsm.group("S")
            a_suffix_num = int(asm.group("N") or "0")
            b_suffix_num = int(bsm.group("N") or "0")
            if a_suffix == b_suffix:
                if b_suffix_num > a_suffix_num:
                    return False
                if a_suffix_num > b_suffix_num:
                    return True
            elif _suffix_gt(a_suffix, b_suffix):
                return True
            else:
                return False
        # More suffixes implies an earlier version,
        # except when the suffix is _p
        if len(self.suffixes) > len(other.suffixes):
            if self.suffixes[len(other.suffixes)].startswith("p"):
                return True
            return False
        if len(self.suffixes) < len(other.suffixes):
            if other.suffixes[len(self.suffixes)].startswith("p"):
                return False
            return True

        # Compare revisions
        if (self.revision or 0) > (other.revision or 0):
            return True
        if (self.revision or 0) < (other.revision or 0):
            return False

        # Equal
        return False


# Fixme: ^= or ~= for partial version matching? (e.g. cargo)
# Or implicit if an operator isn't provided, and wildcards?
# E.g. 2.5 is equivalent to 2.5.*,
# E.g. 2 is equivalent to 2.*, etc.
_range_op_re = r"^(!=|<=|>=|==|<|>)?"


class Operator(Enum):
    Eq = "=="
    Neq = "!="
    Lt = "<"
    Gt = ">"
    Leq = "<="
    Geq = ">="


@dataclass
class VersionRange:
    """
    A comma-separated range of versions which apply simultaneously.

    A version is considered in the range if it matches every component.

    E.g.
    "3.0" is in ">=1.0,!=2.0", but "2.0" and "0.1" are not.

    This class should be treated as immutable
    """

    versions: List[Tuple[Optional[Operator], Version]]

    def __init__(self, inputstr: str, external: bool = False):
        def parse_elem(string: str):
            result = re.split(_range_op_re, string, maxsplit=1)
            if len(result) < 3:
                raise ValueError(f"Version {string} does not begin with an operator!")
            _, op, version = result
            if not op and version.endswith("*"):
                return None, Version(version[:-1], external=external)
            elif not op:
                raise ValueError(
                    f"Version {string} must either begin with an operator, or end with a wildcard"
                )
            else:
                return Operator(op), Version(version, external=external)

        self.versions = [parse_elem(elem) for elem in inputstr.split(",")]

    def __contains__(self, other: Any):
        if isinstance(other, Version):
            for op, version in self.versions:
                if op:
                    if (
                        (op == Operator.Eq and other != version)
                        or (op == Operator.Neq and other == version)
                        or (op == Operator.Lt and other >= version)
                        or (op == Operator.Gt and other <= version)
                        or (op == Operator.Leq and other > version)
                        or (op == Operator.Geq and other < version)
                    ):
                        return False
                else:
                    # Wildcard
                    # e.g. 1*
                    print(
                        other,
                        version,
                        other < version,
                        version.greater(other, ignore_extra_components=True),
                    )
                    if other < version:
                        return False
                    if other.greater(version, ignore_extra_components=True):
                        return False
            return True
        raise NotImplementedError()

    def _get_bounds(self):
        upper_bound = lower_bound = None
        upper_inclusive = lower_inclusive = None

        for op, version in self.versions:
            if op is None:
                raise NotImplementedError()
            if op == Operator.Lt and (upper_bound is None or version > upper_bound):
                upper_bound = version
                upper_inclusive = False
            elif op == Operator.Leq and (upper_bound is None or version >= upper_bound):
                upper_bound = version
                upper_inclusive = True
            if op == Operator.Gt and (lower_bound is None or version < lower_bound):
                lower_bound = version
                lower_inclusive = False
            elif op == Operator.Geq and (lower_bound is None or version <= lower_bound):
                lower_bound = version
                lower_inclusive = True

        if upper_bound and lower_bound:
            # E.g. >2.0,<1.0
            if upper_bound < lower_bound:
                raise ValueError("Upper bound is lower than lower bound!")

            # Equal, but one of the inequalities isn't inclusive
            # E.g. >1.0,<1.0 or >=1.0,<1.0
            if (
                not upper_inclusive or not lower_inclusive
            ) and upper_bound == lower_bound:
                raise ValueError("Upper bound is lower than lower bound!")
        return upper_bound, upper_inclusive, lower_bound, lower_inclusive

    def __and__(self, other: Any):
        if not isinstance(other, VersionRange):
            raise NotImplementedError()

        for op, version in self.versions:
            # If there is an ==, it must be in the other range
            if op == Operator.Eq and version not in other:
                return False
            # If there is a !=, it must not be in the other range
            elif op == Operator.Neq and version in other:
                return False

        # Because range components combine with AND, there are two possibilities
        # 1. Only one inequality, allowing all versions greater or less than a certain version
        # 2. Two inequalities (or more than two where some are redundant), allowing a range between two versions
        try:
            (
                upper_bound,
                upper_inclusive,
                lower_bound,
                lower_inclusive,
            ) = self._get_bounds()
            (
                other_upper_bound,
                other_upper_inclusive,
                other_lower_bound,
                other_lower_inclusive,
            ) = other._get_bounds()
        except ValueError:
            # Contradiction in the bounds, one of the ranges is empty
            return False

        # Cases 1 and 2:
        # E.g. <1.0 and >2.0
        if upper_bound and other_lower_bound and upper_bound < other_lower_bound:
            return False
        # E.g. <1.0 and >1.0
        if (
            upper_bound
            and other_lower_bound
            and (not upper_inclusive or not other_lower_inclusive)
            and upper_bound <= other_lower_bound
        ):
            return False
        # E.g. >2.0 and <1.0
        if lower_bound and other_upper_bound and lower_bound > other_upper_bound:
            return False
        # E.g. >1.0 and <1.0
        if (
            lower_bound
            and other_upper_bound
            and (not lower_inclusive or not other_upper_inclusive)
            and lower_bound >= other_upper_bound
        ):
            return False

        return True

    def __str__(self):
        return ",".join(
            (f"{operator.value}{version}" if operator else f"{version}.*")
            for operator, version in self.versions
        )


def _suffix_gt(a_suffix: str, b_suffix: str) -> bool:
    """Returns true iff a_suffix > b_suffix"""
    suffixes = ["alpha", "beta", "pre", "rc", "p"]
    return suffixes.index(a_suffix) > suffixes.index(b_suffix)
